/*
 *  COPYRIGHT (c) 2009 Chris Johns <chrisj@rtems.org>
 *
 *  COPYRIGHT (c) 1989-2009 On-Line Applications Research Corporation (OAR).
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.rtems.com/license/LICENSE.
 *
 *  $Id: deviceio.c,v 1.21 2009/04/29 08:31:27 ccj Exp $
 */
/**
 * @file
 *
 * @ingroup rtems-ext2fs
 *
 * RTEMS EXT2FS Device Interface.
 *
 * This file contains the set of handlers used to map operations on
 * EXT2 device nodes onto calls to the RTEMS Classic API IO Manager.
 *
 */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include "rtems-ext2fs-private.h"
/*
 * Convert RTEMS status to a UNIX errno
 */
extern int rtems_deviceio_errno(rtems_status_code code);

/*
 *  device_open
 *
 *  This handler maps an open() operation onto rtems_io_open().
 */

static int device_open(
  rtems_libio_t *iop,
  const char    *pathname,
  uint32_t       flag,
  uint32_t       mode
)
{
  rtems_libio_open_close_args_t args;
  ext2_filsys                   fs = rtems_ext2fs_dev(&iop->pathinfo);
  ext2_ino_t                    ino = rtems_ext2fs_get_iop_ino(iop);
  struct ext2_inode             inode;
  errcode_t                     rc;
  int                           major;
  int                           minor;
  rtems_status_code             status;
  
  rc = ext2fs_read_inode(fs, ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("device_open: read inode", rc);

  rc = rtems_ext2fs_inode_device(&inode, &major, &minor);
  if (rc)
    return rtems_ext2fs_remap_error("device_open: inode device", rc);

  iop->data0 = major;
  iop->data1 = (void*)((intptr_t) minor);
  
  args.iop   = iop;
  args.flags = iop->flags;
  args.mode  = mode;

  status = rtems_io_open(major, minor, (void *) &args);
  if (status)
    return rtems_deviceio_errno(status);

  return 0;
}

/*
 *  device_close
 *
 *  This handler maps a close() operation onto rtems_io_close().
 */

static int device_close(
  rtems_libio_t *iop
)
{
  rtems_libio_open_close_args_t args;
  rtems_status_code             status;
  int                           major;
  int                           minor;

  major = (int) iop->data0;
  minor = (intptr_t) iop->data1;

  args.iop   = iop;
  args.flags = 0;
  args.mode  = 0;

  status = rtems_io_close(major, minor, (void *) &args);
  if (status)
    return rtems_deviceio_errno(status);

  return 0;
}

/*
 *  device_read
 *
 *  This handler maps a read() operation onto rtems_io_read().
 */

static ssize_t device_read(
  rtems_libio_t *iop,
  void          *buffer,
  size_t         count
)
{
  rtems_libio_rw_args_t args;
  rtems_status_code     status;
  int                   major;
  int                   minor;

  major = (int) iop->data0;
  minor = (intptr_t) iop->data1;

  args.iop         = iop;
  args.offset      = iop->offset;
  args.buffer      = buffer;
  args.count       = count;
  args.flags       = iop->flags;
  args.bytes_moved = 0;

  status = rtems_io_read(major, minor, (void *) &args);
  if (status)
    return rtems_deviceio_errno(status);

  return (ssize_t) args.bytes_moved;
}

/*
 *  device_write
 *
 *  This handler maps a write() operation onto rtems_io_write().
 */

static ssize_t device_write(
  rtems_libio_t *iop,
  const void    *buffer,
  size_t         count
)
{
  rtems_libio_rw_args_t args;
  rtems_status_code     status;
  int                   major;
  int                   minor;

  major = (int) iop->data0;
  minor = (intptr_t) iop->data1;

  args.iop         = iop;
  args.offset      = iop->offset;
  args.buffer      = (void *) buffer;
  args.count       = count;
  args.flags       = iop->flags;
  args.bytes_moved = 0;

  status = rtems_io_write(major, minor, (void *) &args);
  if (status)
    return rtems_deviceio_errno(status);

  return (ssize_t) args.bytes_moved;
}

/*
 *  device_ioctl
 *
 *  This handler maps an ioctl() operation onto rtems_io_ioctl().
 */

static int device_ioctl(
  rtems_libio_t *iop,
  uint32_t       command,
  void          *buffer
)
{
  rtems_libio_ioctl_args_t args;
  rtems_status_code        status;
  int                      major;
  int                      minor;

  major = (int) iop->data0;
  minor = (intptr_t) iop->data1;

  args.iop     = iop;
  args.command = command;
  args.buffer  = buffer;

  status = rtems_io_control(major, minor, (void *) &args);
  if (status)
    return rtems_deviceio_errno(status);

  return args.ioctl_return;
}

/*
 *  device_lseek
 *
 *  This handler eats all lseek() operations and does not create
 *  an error. It assumes all devices can handle the seek. The
 *  writes fail.
 */

static rtems_off64_t device_lseek(
  rtems_libio_t *iop,
  rtems_off64_t  offset,
  int            whence
)
{
  return offset;
}

/*
 *  device_stat
 *
 *  The rtems_ext2fs_stat() is used.
 */

/*
 *  device_rmnod
 *
 *  The rtems_ext2fs_rmnod() is used.
 */

static int device_ftruncate(
  rtems_libio_t *iop,
  rtems_off64_t  length
)
{
  return 0;
}

/*
 *  Handler table for EXT2FS device nodes
 */

const rtems_filesystem_file_handlers_r rtems_ext2fs_device_handlers = {
  .open_h      = device_open,
  .close_h     = device_close,
  .read_h      = device_read,
  .write_h     = device_write,
  .ioctl_h     = device_ioctl,
  .lseek_h     = device_lseek,
  .fstat_h     = rtems_ext2fs_stat,
  .fchmod_h    = rtems_ext2fs_fchmod,
  .ftruncate_h = device_ftruncate,
  .fpathconf_h = NULL,
  .fsync_h     = NULL,
  .fdatasync_h = NULL,
  .fcntl_h     = NULL,
  .rmnod_h     = rtems_ext2fs_rmnod
};
