/*
 *  COPYRIGHT (c) 2009 Chris Johns <chrisj@rtems.org>
 *
 *  COPYRIGHT (c) 1989-2009 On-Line Applications Research Corporation (OAR).
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.rtems.com/license/LICENSE.
 *
 *  $Id: imfs_directory.c,v 1.23 2009/04/29 08:31:27 ccj Exp $
 */
/**
 * @file
 *
 * @ingroup rtems-ext2fs
 *
 * RTEMS EXTFS Directory Access Routines
 */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <dirent.h>

#include "rtems-ext2fs-private.h"

/*
 *  rtems_ext2fs_dir_open
 *
 *  This rountine will verify that the node being opened as a directory is
 *  in fact a directory node. If it is then the offset into the directory
 *  will be set to 0 to position to the first directory entry.
 */

static int rtems_ext2fs_dir_open(
  rtems_libio_t *iop,
  const char    *pathname,
  uint32_t       flag,
  uint32_t       mode
)
{
  ext2_filsys       fs = rtems_ext2fs_dev(&iop->pathinfo);
  ext2_ino_t        ino = rtems_ext2fs_get_iop_ino(iop);
  struct ext2_inode inode;
  errcode_t         rc;

  rc = ext2fs_read_inode(fs, ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("dir_open: read inode", rc);

  if (!S_ISDIR(inode.i_mode))
    return rtems_ext2fs_remap_error("dir_open: not dir", ENOTDIR);
    
  iop->offset = 0;
  return 0;
}

/*
 *  rtems_ext2fs_dir_close
 *
 *  This routine will be called by the generic close routine to cleanup any
 *  resources that have been allocated for the management of the file
 */

static int rtems_ext2fs_dir_close(
  rtems_libio_t  *iop
)
{
  /*
   *  The generic close routine handles the deallocation of the file control
   *  and associated memory. At present the imfs_dir_close simply
   *  returns a successful completion status.
   */

  return 0;
}

/*
 *  rtems_ext2fs_dir_read
 *
 *  This routine will read the next directory entry based on the directory
 *  offset. The offset should be equal to -n- time the size of an individual
 *  dirent structure. If n is not an integer multiple of the sizeof a
 *  dirent structure, an integer division will be performed to determine
 *  directory entry that will be returned in the buffer. Count should reflect
 *  -m- times the sizeof dirent bytes to be placed in the buffer.
 *  If there are not -m- dirent elements from the current directory position
 *  to the end of the exisiting file, the remaining entries will be placed in
 *  the buffer and the returned value will be equal to -m actual- times the
 *  size of a directory entry.
 */

typedef struct rtems_ext2fs_dirent_s
{
  size_t         position;
  size_t         first_entry;
  size_t         last_entry;
  int            bytes_transferred;
  void*          buffer;
  rtems_libio_t* iop;
} rtems_ext2fs_dirent;

static int rtems_ext2fs_dir_find(struct ext2_dir_entry* de,
                                 int                    offset,
                                 int                    blocksize,
                                 char*                  buf,
                                 void*                  priv_data)
{
  rtems_ext2fs_dirent* rde = priv_data;
	struct dirent        dirent;

  if (rde->position >= rde->first_entry)
  {
    /*
     * Move the entry to the return buffer.
     */
    memset(&dirent, 0, sizeof(struct dirent));
    dirent.d_off = offset;
    dirent.d_reclen = sizeof(struct dirent);
    dirent.d_ino = de->inode;
    dirent.d_namlen = de->name_len & 0xff;
    memcpy(dirent.d_name, de->name, dirent.d_namlen);
    dirent.d_name[dirent.d_namlen] = '\0';
  
    memcpy(rde->buffer + rde->bytes_transferred,
           &dirent, sizeof(struct dirent));
    rde->iop->offset = rde->iop->offset + sizeof(struct dirent);
    rde->bytes_transferred += sizeof(struct dirent);

  }
  
  rde->position += sizeof(struct dirent);
    
  return rde->position >= rde->last_entry ? BLOCK_ABORT : 0;
}

static ssize_t rtems_ext2fs_dir_read(
  rtems_libio_t  *iop,
  void           *buffer,
  size_t          count
)
{
  ext2_filsys         fs = rtems_ext2fs_dev(&iop->pathinfo);
  ext2_ino_t          ino = rtems_ext2fs_get_iop_ino(iop);
  struct ext2_inode   inode;
  rtems_ext2fs_dirent de;
  errcode_t           rc;

  rc = ext2fs_read_inode(fs, ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("dir_read: read inode", rc);

  de.position = 0;
  de.first_entry = iop->offset;
  /* protect against using sizes that are not exact multiples of the */
  /* -dirent- size. These could result in unexpected results          */
  de.last_entry = de.first_entry + (count / sizeof(struct dirent)) * sizeof(struct dirent);
  de.bytes_transferred = 0;
  de.buffer = buffer;
  de.iop = iop;

  rc = ext2fs_dir_iterate(fs, ino, 0, NULL, rtems_ext2fs_dir_find, &de);
	if (rc)
    return rtems_ext2fs_remap_error("dir_read: dir iterate", rc);
  
   /* Success */
  return de.bytes_transferred;
}

/*
 *  rtems_ext2fs_dir_lseek
 *
 *  This routine will behave in one of three ways based on the state of
 *  argument whence. Based on the state of its value the offset argument will
 *  be interpreted using one of the following methods:
 *
 *     SEEK_SET - offset is the absolute byte offset from the start of the
 *                logical start of the dirent sequence that represents the
 *                directory
 *     SEEK_CUR - offset is used as the relative byte offset from the current
 *                directory position index held in the iop structure
 *     SEEK_END - N/A --> This will cause an assert.
 */

static rtems_off64_t rtems_ext2fs_dir_lseek(
  rtems_libio_t  *iop,
  rtems_off64_t   offset,
  int             whence
)
{
  switch( whence )
  {
     case SEEK_SET:   /* absolute move from the start of the file */
     case SEEK_CUR:   /* relative move */
       iop->offset =
         (iop->offset/sizeof(struct dirent)) * sizeof(struct dirent);
        break;

     case SEEK_END:   /* Movement past the end of the directory via lseek */
                      /* is not a permitted operation                     */
      default:
        return rtems_ext2fs_remap_error("dir_lseek: bad whence", EINVAL);
        break;
  }
  return 0;
}

/*
 * rtems_ext2fs_dir_fcntl
 *
 * The following routine does a fcntl on an node.
 */

int rtems_ext2fs_dir_fcntl(
  int            cmd,
  rtems_libio_t *iop
)
{
  return 0;
}

/*
 *  rtems_extfs_dir_rmnod
 *
 *  This routine is available from the optable to remove a node
 *  from the EXT2 file system.
 */
typedef struct rtems_ext2fs_dir_empty_s
{
  ext2_ino_t parent;
  bool       empty;
} rtems_ext2fs_dir_empty;

static int rtems_ext2fs_dir_empty_checker(struct ext2_dir_entry* de,
                                          int                    offset,
                                          int                    blocksize,
                                          char*                  buf,
                                          void*                  priv_data)
{
  rtems_ext2fs_dir_empty* rde = priv_data;
  int                     len = de->name_len & 0xff;
  
#if RTEMS_EXT2FS_DEBUG_DIR_RMNOD
  printf("rtems-ext2fs: dir-rmnod: empty-checker: ino:%i name:%s len:%d\n",
         de->inode, de->name, len);
#endif
  
  if (!de->inode || ((len == 1) && (de->name[0] == '.')))
    return 0;
  if ((len == 2) && (de->name[0] == '.') && (de->name[1] == '.'))
  {
#if RTEMS_EXT2FS_DEBUG_DIR_RMNOD
  printf("rtems-ext2fs: dir-rmnod: empty-checker: parent found\n");
#endif
    rde->parent = de->inode;
    return 0;
  }
  
  rde->empty = false;
  return 0;
}

static int rtems_ext2fs_dir_rmnod(
  rtems_filesystem_location_info_t      *parent_pathloc, /* IN */
  rtems_filesystem_location_info_t      *pathloc         /* IN */
)
{
  ext2_filsys            fs = rtems_ext2fs_dev(pathloc);
  ext2_ino_t             parent_ino = rtems_ext2fs_get_pathloc_ino(parent_pathloc);
  ext2_ino_t             ino = rtems_ext2fs_get_pathloc_ino(pathloc);
  struct ext2_inode      inode;
  rtems_ext2fs_dir_empty de;
  errcode_t              rc;

#if RTEMS_EXT2FS_DEBUG_DIR_RMNOD
  printf("rtems-ext2fs: dir-rmnod: parent:%i ino:%i\n", parent_ino, ino);
#endif
  
  if (ino == EXT2_ROOT_INO)
    return rtems_ext2fs_remap_error("dir_rmnod: root inode", EBUSY);

  rc = ext2fs_read_inode(fs, ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("dir_rmnod: read inode", rc);

  if (!LINUX_S_ISDIR(inode.i_mode))
    return rtems_ext2fs_remap_error("dir_rmnod: not dir", ENOTDIR);

  de.parent = 0;
  de.empty = true;
   
  rc = ext2fs_dir_iterate(fs, ino, 0, NULL, rtems_ext2fs_dir_empty_checker, &de);
	if (rc)
    return rtems_ext2fs_remap_error("dir_rmnod: dir iterate", rc);

  if (!de.empty)
    return rtems_ext2fs_remap_error("dir_rmnod: not empty", ENOTEMPTY);

  if (de.parent == 0)
    return rtems_ext2fs_remap_error("dir_rmnod: no parent", EIO);

  if (parent_ino != de.parent)
    return rtems_ext2fs_remap_error("dir_rmnod: not parent", EIO);
    
	rc = ext2fs_unlink(fs, de.parent, NULL, ino, 0);
	if (rc)
    return rtems_ext2fs_remap_error("dir_rmnod: unlink", rc);

	rc = rtems_ext2fs_unlinkbyinode(fs, ino, &inode);
	if (rc)
    return rtems_ext2fs_remap_error("dir_rmnod: unlink by inode", rc);

  rc = ext2fs_read_inode(fs, de.parent, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("dir_rmnod: read inode", rc);

  if (inode.i_links_count > 1)
    inode.i_links_count--;
    
  rc = ext2fs_write_inode(fs, de.parent, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("dir_rmnod: write inode", rc);

  rc = ext2fs_write_bitmaps(fs);
	if (rc)
    return rtems_ext2fs_remap_error("dir_rmnod: write bitmaps", rc);
  
  return 0;
}

/*
 *  Set of operations handlers for operations on directories.
 */

const rtems_filesystem_file_handlers_r rtems_ext2fs_dir_handlers = {
  .open_h      = rtems_ext2fs_dir_open,
  .close_h     = rtems_ext2fs_dir_close,
  .read_h      = rtems_ext2fs_dir_read,
  .write_h     = NULL,
  .ioctl_h     = NULL,
  .lseek_h     = rtems_ext2fs_dir_lseek,
  .fstat_h     = rtems_ext2fs_stat,
  .fchmod_h    = rtems_ext2fs_fchmod,
  .ftruncate_h = NULL,
  .fpathconf_h = NULL,
  .fsync_h     = NULL,
  .fdatasync_h = rtems_ext2fs_fdatasync,
  .fcntl_h     = rtems_ext2fs_fcntl,
  .rmnod_h     = rtems_ext2fs_dir_rmnod
};
