/*
 *  COPYRIGHT (c) 2009 Chris Johns <chrisj@rtems.org>
 *
 *  COPYRIGHT (c) 1989-2009 On-Line Applications Research Corporation (OAR).
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.rtems.com/license/LICENSE.
 *
 *  $Id$
 */
/**
 * @file
 *
 * @ingroup rtems-ext2fs
 *
 * Set of utility functions to support EXT2FS on RTEMS.
 *
 */

#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>

#include <rtems/seterr.h>
#include "rtems-ext2fs-private.h"

/*
 * Convert the mode to EXT2 flags. See ext2fs_link in the info for
 * libext2fs for details.
 */
int rtems_ext2fs_imodeflags(mode_t mode)
{
  int imode = (mode & 0777);
  if (S_ISREG(mode))
    imode |= LINUX_S_IFREG;
  else if (S_ISDIR(mode))
    imode |= LINUX_S_IFDIR;
  else if (S_ISCHR(mode))
    imode |= LINUX_S_IFCHR;
  else if (S_ISBLK(mode))
    imode |= LINUX_S_IFBLK;
  else if (S_ISFIFO(mode))
    imode |= LINUX_S_IFIFO;
  else if (S_ISSOCK(mode))
    imode |= LINUX_S_IFSOCK;
  else if (S_ISLNK(mode))
    imode |= LINUX_S_IFLNK;
  return imode;
}

/*
 * Convert the EXT2FS flags to mode flags
 */
mode_t rtems_ext2fs_modeflags(int imode)
{
  mode_t mode;
  if (LINUX_S_ISREG(imode))
    mode = S_IFREG;
  else if (LINUX_S_ISDIR(imode))
    mode = S_IFDIR;
  else if (LINUX_S_ISCHR(imode))
    mode = S_IFCHR;
  else if (LINUX_S_ISBLK(imode))
    mode = S_IFBLK;
  else if (LINUX_S_ISFIFO(imode))
    mode = S_IFIFO;
  else if (LINUX_S_ISSOCK(imode))
    mode = S_IFSOCK;
  else if (LINUX_S_ISLNK(imode))
    mode = S_IFLNK;
  return mode | (imode & 0777);
}

/*
 * Read an ino and inode for a specific path.
 */
int rtems_ext2fs_read_inode(ext2_filsys        fs,
                            const char*        path,
                            ext2_ino_t         cwd,
                            ext2_ino_t*        ino,
                            struct ext2_inode* inode)
{
  errcode_t rc;
  rc = ext2fs_namei(fs, EXT2_ROOT_INO, cwd, path, ino);
  if (rc)
    return ENOENT;
  rc = ext2fs_read_inode(fs, *ino, inode);
  return rc;
}

/*
 * Read parent ino and inode for a specific ino.
 */
int rtems_ext2fs_read_parent_inode(ext2_filsys        fs,
                                   ext2_ino_t         ino,
                                   ext2_ino_t*        pino,
                                   struct ext2_inode* pinode)
{
  errcode_t rc;
  rc = ext2fs_namei(fs, EXT2_ROOT_INO, ino, "..", pino);
  if (rc)
    return ENOENT;
  rc = ext2fs_read_inode(fs, *pino, pinode);
  return rc;
}

/*
 * Extract the device number from a device inode.
 */
int rtems_ext2fs_inode_device(struct ext2_inode* inode,
                              int*               major,
                              int*               minor)
{
  if (!S_ISCHR(inode->i_mode) && !S_ISBLK(inode->i_mode))
    return ENODEV;

  if (inode->i_block[0])
  {
    *major = (inode->i_block[0] >> 8) & 255;
    *minor = inode->i_block[0] & 255;
  }
  else
  {
    *major = (inode->i_block[1] & 0xfff00) >> 8;
    *minor = ((inode->i_block[1] & 0xff) |
              ((inode->i_block[1] >> 12) & 0xfff00));
  }

  return 0;
}

static int release_blocks(ext2_filsys fs, blk_t* block, int blockcnt, void* private)
{
  ext2fs_block_alloc_stats(fs, *block, -1);
  return 0;
}

int rtems_ext2fs_unlinkbyinode(ext2_filsys        fs,
                               ext2_ino_t         ino,
                               struct ext2_inode* inode)
{
  errcode_t rc;

  inode->i_links_count = 0;
  inode->i_dtime = time(NULL);

  rc = ext2fs_write_inode(fs, ino, inode);
  if (rc)
    return EIO;

  if (ext2fs_inode_has_valid_blocks(inode))
  {
    ext2fs_block_iterate(fs, ino, 0, NULL, release_blocks, NULL);
    ext2fs_inode_alloc_stats2(fs, ino, -1, LINUX_S_ISDIR(inode->i_mode));
  }

  return 0;
}

/*
 *  rtems_ext2fs_evaluate_permission
 *
 *  The following routine evaluates that we have permission
 *  to do flags on the node.
 */

int rtems_ext2fs_evaluate_permission(struct ext2_inode* inode, int flags)
{
  uid_t st_uid;
  gid_t st_gid;
  int   flags_to_test;

#if defined(RTEMS_POSIX_API)
  st_uid = geteuid();
  st_gid = getegid();
#else
  st_uid = inode_uid(*inode);
  st_gid = inode_gid(*inode);
#endif

  /*
   * Check if I am owner or a group member or someone else.
   */
  flags_to_test = flags;

  if ((st_uid == 0) || (st_uid == inode_uid(*inode)))
    flags_to_test |= flags << 6;
  if ((st_uid == 0) || (st_gid == inode_gid(*inode)))
    flags_to_test |= flags << 3;
  else
    /* must be other - already set above */;

#if RTEMS_EXT2FS_DEBUG_EVAL_PERMS
  printf ("rtems-ext2fs: eval-perms: uid=%d gid=%d iuid=%d igid=%d " \
          "flags=%o flags_to_test=%o mode=%o (%o)\n",
          st_uid, st_gid, inode->i_uid, inode->i_gid,
          flags, flags_to_test, inode->i_mode & 0777,
          flags_to_test & (inode->i_mode & 0777));
#endif
  
  /*
   * If all of the flags are set we have permission
   * to do this.
   */
  if ((flags_to_test & (inode->i_mode & 0777)) != 0)
    return 1;

#if RTEMS_EXT2FS_DEBUG_EVAL_PERMS
  printf("rtems-ext2fs: eval-perms: perms failed\n");
#endif
  return 0;
}

/*
 *  rtems_ext2fs_set_handlers
 *
 *  The following sets the handlers based on the type of inode.
 */

int rtems_ext2fs_set_handlers(rtems_filesystem_location_info_t* loc,
                              struct ext2_inode*                inode)
{
  loc->handlers = NULL;
  if (LINUX_S_ISDIR(inode->i_mode))
    loc->handlers = rtems_ext2fs_handlers(dir);
  else if (LINUX_S_ISCHR(inode->i_mode) || LINUX_S_ISBLK(inode->i_mode))
    loc->handlers = rtems_ext2fs_handlers(device);
  else if (LINUX_S_ISLNK(inode->i_mode))
    loc->handlers = rtems_ext2fs_handlers(link);
  else if (LINUX_S_ISREG(inode->i_mode))
    loc->handlers = rtems_ext2fs_handlers(file);
  return 0;
}

/*
 * Remap the error codes.
 */
int rtems_ext2fs_remap_error(const char* what, int rc)
{
  if (rc <= 0)
    return rc;

#if RTEMS_EXT2FS_DEBUG_ERROR_MSGS
  {
    const char* message = error_message(rc);
    printf ("rtems-ext2fs: error: %s: %s\n", what, message);
  }
#endif
  
  if (rc >= ext2_err_base)
  {
    if (rc == EXT2_ET_NO_MEMORY)
      rc = ENOMEM;
    else
      rc = EIO;
  }
  
  rtems_set_errno_and_return_minus_one(rc);
}
