/*
 *  COPYRIGHT (c) 2009 Chris Johns <chrisj@rtems.org>
 *
 *  COPYRIGHT (c) 1989-2009.  On-Line Applications Research Corporation (OAR).
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 *  http://www.rtems.com/license/LICENSE.
 *
 *  $Id$
 */
/**
 * @file
 *
 * @ingroup rtems-ext2fs
 *
 * RTEMS EXT2FS File System Interface.
 */


#if HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>

#include "rtems-ext2fs-private.h"

#define RTEMS_LIBIO_PERMS_RX (RTEMS_LIBIO_PERMS_SEARCH | RTEMS_LIBIO_PERMS_READ)
#define RTEMS_LIBIO_PERMS_WX (RTEMS_LIBIO_PERMS_SEARCH | RTEMS_LIBIO_PERMS_WRITE)

/*
 *  rtems_ext2fs_eval_path
 *
 *  The following routine evaluates the path to a node that wishes to be
 *  accessed.  The pathloc is returned with the ino to the node to be accessed.
 *
 *  The routine starts from the root stripping away any leading path separators
 *  breaking the path up into the node names and checking an inode exists for
 *  that node name. Permissions are checked to insure access to the node is
 *  allowed. A path to a node must be accessable all the way even if the end
 *  result is directly accessable. As a user on Linux try "ls /root/../tmp" and
 *  you will see if fails.
 *
 *  The whole process is complicated by crossmount paths where we head down
 *  into this file system only to return to the top and out to a another
 *  mounted file system. For example we are mounted on '/e' and the user enters
 *  "ls /e/a/b/../../dev". We need to head down then back up.
 *  
 */

int rtems_ext2fs_eval_path(
  const char                        *path,         /* IN     */
  int                                pathlen,      /* IN     */
  int                                flags,        /* IN     */
  rtems_filesystem_location_info_t  *pathloc       /* IN/OUT */
)
{
  struct ext2_inode inode;
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(pathloc);
  char*             node;
  int               stripped;
  int               n;
  int               rc;

#if RTEMS_EXT2FS_DEBUG_EVAL_PATH
  printf("rtems-ext2fs: eval-path: in: path:%s pathlen:%i ino:%d\n", path, pathlen, ino);
#endif
  
  node = malloc(RTEMS_EXT2FS_MAX_NAME);
  if (!node)
    return rtems_ext2fs_remap_error("eval_path: node memory", ENOMEM);

  /*
   * Read the entering inode so we know we can enter it.
   */
  rc = ext2fs_read_inode(rtems_ext2fs_dev(pathloc), ino, &inode);
  if (rc)
  {
    free(node);
    return rtems_ext2fs_remap_error("eval_path: reading inode", rc);
  }
  
  /*
   * Eat any separators at start of the path.
   */
  stripped = rtems_filesystem_prefix_separators(path, pathlen);
  path += stripped;
  pathlen -= stripped;

  while (true)
  {
    /*
     * Is this the end of the pathname we where given ?
     */
    if ((*path == '\0') || (pathlen == 0))
      break;

    /*
     * If a directory the execute bit must be set for us to enter.
     */
    if (LINUX_S_ISDIR(inode.i_mode) &&
        !rtems_ext2fs_evaluate_permission(&inode, RTEMS_LIBIO_PERMS_SEARCH))
    {
      free(node);
      return rtems_ext2fs_remap_error("eval_path: eval perms", EACCES);
    }

    /*
     * Extract the node name we will look for this time around.
     */
    n = 0;
    while (!rtems_filesystem_is_separator(*path) &&
           (*path != '\0') && pathlen &&
           (n < (RTEMS_EXT2FS_MAX_NAME - 1)))
    {
      node[n] = *path;
      n++;
      node[n] = '\0';
      path++;
      pathlen--;
    }

    /*
     * Eat any separators at start of the path.
     */
    stripped = rtems_filesystem_prefix_separators(path, pathlen);
    path += stripped;
    pathlen -= stripped;
    n += stripped;
    
    if (rtems_extfs_current_dir(node))
    {
      if (*path)
        continue;
      break;
    }
    
    if (rtems_extfs_parent_dir(node))
    {
      /*
       * If we are at root inode of the file system we have a crossmount path.
       */
      if (ino == EXT2_ROOT_INO)
      {
#if RTEMS_EXT2FS_DEBUG_EVAL_PATH
        printf("rtems-ext2fs: eval-path: crossmount: path:%s (%d)\n",
               path - n, pathlen + n);
#endif
        free(node);
        *pathloc = pathloc->mt_entry->mt_point_node;
        return (*pathloc->ops->evalpath_h)(path - n, pathlen + n,
                                           flags, pathloc);
      }

      /*
       * We need to find the parent of this node.
       */
      rc = rtems_ext2fs_read_parent_inode(rtems_ext2fs_dev(pathloc),
                                          ino, &ino, &inode);
      if (rc)
      {
        free(node);
        return rtems_ext2fs_remap_error("eval_path: read parent inode", rc);
      }
#if RTEMS_EXT2FS_DEBUG_EVAL_PATH
      printf("rtems-ext2fs: eval-path: parent: ino:%d\n", ino);
#endif
    }
    else
    {
      /*
       * Read the inode so we know it exists and what type it is.
       */
      rc = rtems_ext2fs_read_inode(rtems_ext2fs_dev(pathloc),
                                   node, ino, &ino, &inode);
      if (rc)
      {
        free(node);
        return rtems_ext2fs_remap_error("eval_path: read node", rc);
      }    
#if RTEMS_EXT2FS_DEBUG_EVAL_PATH
      printf("rtems-ext2fs: eval-path: down: path:%s ino:%d\n", node, ino);
#endif
    }
  }
  
  free(node);

  rtems_ext2fs_set_pathloc_ino(pathloc, ino);
  rtems_ext2fs_set_handlers(pathloc, &inode);

#if RTEMS_EXT2FS_DEBUG_EVAL_PATH
  printf("rtems-ext2fs: eval-path: ino:%d\n", ino);
#endif
  return 0;
}

/*
 *  rtems_ext2fs_eval_for_make
 *
 *  The following routine evaluate path for a new node to be created.
 *  pathloc is returned with a pointer to the parent of the new node.
 *  name is returned with a pointer to the first character in the
 *  new node name.  The parent node is verified to be a directory.
 */

int rtems_ext2fs_eval_for_make(
   const char                         *path,       /* IN     */
   rtems_filesystem_location_info_t   *pathloc,    /* IN/OUT */
   const char                        **name        /* OUT    */
)
{
  struct ext2_inode inode;
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(pathloc);
  char*             node;
  int               stripped;
  int               n;
  int               rc;

#if RTEMS_EXT2FS_DEBUG_EVAL_FOR_MAKE
  printf("rtems-ext2fs: eval-for-make: path:%s ino:%d\n", path, ino);
#endif
  
  node = malloc(RTEMS_EXT2FS_MAX_NAME);
  if (!node)
    return rtems_ext2fs_remap_error("eval_for_make: no memory", ENOMEM);

  *name = path + strlen(path);

  while (*name != path)
  {
    (*name)--;
    if (rtems_filesystem_is_separator(**name))
    {
      (*name)++;
      break;
    }
  }
      
  /*
   * Eat any separators at start of the path.
   */
  stripped = rtems_filesystem_prefix_separators(path, strlen(path));
  path += stripped;

  /*
   * Read the entering inode so we know we can enter it.
   */
  rc = ext2fs_read_inode(rtems_ext2fs_dev(pathloc), ino, &inode);
  if (rc)
  {
    free(node);
    return rtems_ext2fs_remap_error("eval_for_make: read ino", rc);
  }
  
  while (true)
  {
    /*
     * If a directory the execute bit must be set for us to enter.
     */
    if (LINUX_S_ISDIR(inode.i_mode) &&
        !rtems_ext2fs_evaluate_permission(&inode, RTEMS_LIBIO_PERMS_SEARCH))
    {
      free(node);
      return rtems_ext2fs_remap_error("eval_for_make: eval perms", EACCES);
    }
    
    /*
     * Is this the end of the pathname we where given ?
     */
    if (path == *name)
      break;
    
    /*
     * Extract the node name we will look for this time around.
     */
    n = 0;
    while (!rtems_filesystem_is_separator(*path) &&
           (*path != '\0') &&
           (n < (RTEMS_EXT2FS_MAX_NAME - 1)))
    {
      node[n] = *path;
      n++;
      node[n] = '\0';
      path++;
    }

    /*
     * Eat any separators at start of the path.
     */
    stripped = rtems_filesystem_prefix_separators(path, strlen(path));
    path += stripped;
    n += stripped;


    if (rtems_extfs_current_dir(node))
    {
      if (*path)
        continue;
      break;
    }
     
    if (rtems_extfs_parent_dir(path))
    {
      /*
       * If we are at root inode of the file system we have a crossmount path.
       */
      if (ino == EXT2_ROOT_INO)
      {
#if RTEMS_EXT2FS_DEBUG_EVAL_FOR_MAKE
        printf("rtems-ext2fs: eval-for-make: crossmount: path:%s\n", path - n);
#endif
        free(node);
        *pathloc = pathloc->mt_entry->mt_point_node;
        return (*pathloc->ops->evalformake_h)(path - n, pathloc, name);
      }

      /*
       * If not a directory and up return. We cannot change dir from a
       * regular file or device node.
       */
      if (!LINUX_S_ISDIR(inode.i_mode))
      {
        free(node);
        return rtems_ext2fs_remap_error("eval_for_make: not dir", ENOTSUP);
      }
      
      /*
       * We need to find the parent of this node.
       */
      rc = rtems_ext2fs_read_parent_inode(rtems_ext2fs_dev(pathloc),
                                          ino, &ino, &inode);
      if (rc)
      {
        free(node);
        return rtems_ext2fs_remap_error("eval_for_make: read parent inode",
                                        rc);
      }
#if RTEMS_EXT2FS_DEBUG_EVAL_FOR_MAKE
      printf("rtems-ext2fs: eval-for-make: parent: ino:%d\n", ino);
#endif
    }
    else
    {
      /*
       * Read the inode so we know it exists and what type it is.
       */
      rc = rtems_ext2fs_read_inode(rtems_ext2fs_dev(pathloc),
                                   node, ino, &ino, &inode);
      if (rc)
      {
        free(node);
        return rtems_ext2fs_remap_error("eval_for_make: reading inode", rc);
      }
#if RTEMS_EXT2FS_DEBUG_EVAL_FOR_MAKE
      printf("rtems-ext2fs: eval-for-make: down: path:%s ino:%d\n", node, ino);
#endif
    }
  }

  free(node);
  
  if (!LINUX_S_ISDIR(inode.i_mode))
    return rtems_ext2fs_remap_error("eval_for_make: not dir", ENOTDIR);

  if (!rtems_ext2fs_evaluate_permission(&inode, RTEMS_LIBIO_PERMS_WX))
    return rtems_ext2fs_remap_error("eval_for_make: cannot write", EACCES);

  /*
   * Set the parent ino in the path location.
   */

  rtems_ext2fs_set_pathloc_ino(pathloc, ino);
  rtems_ext2fs_set_handlers(pathloc, &inode);

#if RTEMS_EXT2FS_DEBUG_EVAL_FOR_MAKE
  printf("rtems-ext2fs: eval-for-make: parent ino:%d name:%s\n", ino, *name);
#endif
  return 0;
}

/*
 * rtems_ext2fs_link
 *
 * The following rouine creates a new link node under parent with the
 * name given in name.  The link node is set to point to the node at
 * to_loc.
 */

int rtems_ext2fs_link(
  rtems_filesystem_location_info_t  *to_loc,      /* IN */
  rtems_filesystem_location_info_t  *parent_loc,  /* IN */
  const char                        *name         /* IN */
)
{
  ext2_filsys       fs = rtems_ext2fs_dev(to_loc);
  ext2_ino_t        to_ino = rtems_ext2fs_get_pathloc_ino(to_loc);
  struct ext2_inode to_inode;
  ext2_ino_t        parent_ino = rtems_ext2fs_get_pathloc_ino(parent_loc);
  struct ext2_inode parent_inode;
  errcode_t         rc;

  rc = ext2fs_read_inode(fs, to_ino, &to_inode);
  if (rc)
    return rtems_ext2fs_remap_error("link: read inode", rc);

  rc = ext2fs_read_inode(fs, parent_ino, &parent_inode);
  if (rc)
    return rtems_ext2fs_remap_error("link: read parent inode", rc);

	do
  { 
		rc = ext2fs_link(fs, parent_ino, name, to_ino,
                     rtems_ext2fs_imodeflags(to_inode.i_mode));
		if (rc == EXT2_ET_DIR_NO_SPACE)
    {
			if (ext2fs_expand_dir(fs, to_ino))
        return rtems_ext2fs_remap_error("link: expand dir", ENOSPC);
		}
	}
  while (rc == EXT2_ET_DIR_NO_SPACE);

	if (rc)
    return rtems_ext2fs_remap_error("link: linking", rc);

	parent_inode.i_mtime = parent_inode.i_ctime = to_inode.i_ctime = time(NULL);

	rc = ext2fs_write_inode(fs, parent_ino, &parent_inode);
	if (rc)
    return rtems_ext2fs_remap_error("link: write parent inode", rc);

  to_inode.i_links_count += 1;
	rc = ext2fs_write_inode(fs, to_ino, &to_inode);
	if (rc)
    return rtems_ext2fs_remap_error("link: write link inode", rc);

  rc = ext2fs_write_bitmaps(fs);
	if (rc)
    return rtems_ext2fs_remap_error("link: write bitmaps", rc);
  
	return 0;
}

/*
 * rtems_ext2fs_unlink
 *
 * Routine to remove a link node from the tree.
 */

int rtems_ext2fs_unlink(
  rtems_filesystem_location_info_t  *parent_loc,  /* IN */
  rtems_filesystem_location_info_t  *loc          /* IN */
)
{
  ext2_ino_t        parent_ino = rtems_ext2fs_get_pathloc_ino(parent_loc);
  struct ext2_inode parent_inode;
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(loc);
  struct ext2_inode inode;
  errcode_t         rc;
  
#if RTEMS_EXT2FS_DEBUG_UNLINK
  printf("rtems-ext2fs: unlink: parent:%i ino:%d\n", parent_ino, ino);
#endif
  
  rc = ext2fs_read_inode(rtems_ext2fs_dev(loc), ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("unlink: read inode", rc);

  if (LINUX_S_ISDIR(inode.i_mode))
    return rtems_ext2fs_remap_error("unlink: inode is dir", EIO);

  rc = ext2fs_read_inode(rtems_ext2fs_dev(loc),
                         parent_ino, &parent_inode);
  if (rc)
    return rtems_ext2fs_remap_error("unlink: read parent inode", rc);

  if (inode.i_links_count > 0)
    inode.i_links_count--;

  rc = ext2fs_unlink(rtems_ext2fs_dev(loc), parent_ino, NULL, ino, 0);
  if (rc)
    return rtems_ext2fs_remap_error("unlink: unlink inode", rc);

#if RTEMS_EXT2FS_DEBUG_UNLINK
  printf("rtems-ext2fs: unlink: links::%i\n", inode.i_links_count);
#endif
  
  if (inode.i_links_count < 1)
  { 
    parent_inode.i_ctime = parent_inode.i_mtime = time(NULL);
  
    rc = rtems_ext2fs_unlinkbyinode(rtems_ext2fs_dev(loc), ino, &inode);
    if (rc)
      return rtems_ext2fs_remap_error("unlink: unlinking by inode", rc);

    rc = ext2fs_write_inode(rtems_ext2fs_dev(loc), parent_ino, &parent_inode);
    if (rc)
      return rtems_ext2fs_remap_error("unlink: write parent inode", rc);
  }
  else
  {
    inode.i_ctime = time(NULL);
    rc = ext2fs_write_inode(rtems_ext2fs_dev(loc), ino, &inode);
    if (rc)
      return rtems_ext2fs_remap_error("unlink: write inode", rc);
  }

  rc = ext2fs_write_bitmaps(rtems_ext2fs_dev(loc));
	if (rc)
    return rtems_ext2fs_remap_error("unlink: write bitmaps", rc);
  
  return 0;
}

/*
 * rtems_ext2fs_node_type
 *
 * The following verifies that returns the type of node that the
 * loc refers to.
 */

rtems_filesystem_node_types_t rtems_ext2fs_node_type(
   rtems_filesystem_location_info_t *pathloc         /* IN */
)
{
  rtems_filesystem_node_types_t type;
  ext2_ino_t                    ino = rtems_ext2fs_get_pathloc_ino(pathloc);
  struct ext2_inode             inode;
  errcode_t                     rc;
  
  rc = ext2fs_read_inode(rtems_ext2fs_dev(pathloc), ino, &inode);
  if (rc) 
    return rtems_ext2fs_remap_error("node_type: read inode", rc);

  /*
   * Do not return RTEMS_FILESYSTEM_HARD_LINK or RTEMS_FILESYSTEM_SYM_LINK
   * because this would result in an eval link which does not make sense in the
   * case of the ext2 file system. The ext2fs_namei call will follow links as
   * you would expect and ends at a link. All directory nodes are links to
   * the inode of the file. A link such as a HARD link is actual the path to
   * the regular file, directory, device etc's inode. Links to inodes can be
   * considered "the real" one, they are all links.
   */
  if (LINUX_S_ISDIR(inode.i_mode))
    type = RTEMS_FILESYSTEM_DIRECTORY;
  else if (LINUX_S_ISREG(inode.i_mode) || LINUX_S_ISLNK(inode.i_mode))
    type = RTEMS_FILESYSTEM_MEMORY_FILE;
  else
    type = RTEMS_FILESYSTEM_DEVICE;
  return type;
}

/*
 * rtems_ext2fs_chown
 *
 * This routine is the implementation of the chown() system
 * call for the EXT2FS.
 *
 */

static int rtems_ext2fs_chown(
  rtems_filesystem_location_info_t *pathloc,       /* IN */
  uid_t                             owner,         /* IN */
  gid_t                             group          /* IN */
)
{
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(pathloc);
  struct ext2_inode inode;
#if defined(RTEMS_POSIX_API)
  uid_t             uid;
#endif
  errcode_t         rc;
  
  rc = ext2fs_read_inode(rtems_ext2fs_dev(pathloc), ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("chown: read inode", rc);

  /*
   *  Verify I am the owner of the node or the super user.
   */

#if defined(RTEMS_POSIX_API)
  uid = geteuid();

  if ((uid != inode.i_uid) && (uid != 0))
    return rtems_ext2fs_remap_error("chown: not able", EPERM);
#endif

  inode.i_uid = owner;
  inode.i_gid = group;

  inode.i_ctime = time(NULL);

  rc = ext2fs_write_inode(rtems_ext2fs_dev(pathloc), ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("chown: write inode", rc);

  return 0;
}

/*
 *  rtems_ext2fs_freenodinfo
 *
 *  This routine is the EXT2 free node handler for the file system
 *  operations table.
 *
 *  Will do something here when I know what it is suppose to do.
 */

int rtems_ext2fs_freenodinfo(
 rtems_filesystem_location_info_t *pathloc       /* IN */
)
{
  return 0;
}

/*
 * rtems_ext2fs_shutdown
 *
 * Shutdown the file system.
 */
int rtems_ext2fs_shutdown(
  rtems_filesystem_mount_table_entry_t *mt_entry
)
{
  ext2_filsys fs = mt_entry->fs_info;
  errcode_t   rc;
  ext2fs_flush(fs);
  ext2fs_write_bitmaps(fs);
  rc = ext2fs_close(fs);
  return rtems_ext2fs_remap_error("shutdown: close", rc);
}

/*
 * rtems_ext2fs_utime
 *
 * This routine is the implementation of the utime() system
 * call for the ext2fs.
 */

int rtems_ext2fs_utime(
  rtems_filesystem_location_info_t  *pathloc,       /* IN */
  time_t                             actime,        /* IN */
  time_t                             modtime        /* IN */
)
{
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(pathloc);
  struct ext2_inode inode;
  errcode_t         rc;

  rc = ext2fs_read_inode(rtems_ext2fs_dev(pathloc), ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("utime: read inode", rc);

  inode.i_atime = actime;
  inode.i_mtime = modtime;

  rc = ext2fs_write_inode(rtems_ext2fs_dev(pathloc), ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("utime: write inode", rc);

  return 0;
}

/*
 * rtems_ext2fs_symlink

 * The following rouine creates a new symbolic link node under parent
 * with the name given in name.  The node is set to point to the node at
 * to_loc.
 */

int rtems_ext2fs_symlink(
  rtems_filesystem_location_info_t *parent_loc,
  const char                       *link_name,
  const char                       *node_name
)
{
  return 0;
}

/*
 * rtems_ext2fs_readlink
 * 
 * The following rouine puts the symblic links destination name into
 * buf.
 */

int rtems_ext2fs_readlink(
  rtems_filesystem_location_info_t *loc,
  char                             *buf,         /* OUT */
  size_t                            bufsize
)
{
  ext2_filsys       fs = rtems_ext2fs_dev(loc);
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(loc);
  struct ext2_inode inode;
  uint8_t*          block = NULL;
  char*             name;
  int               length;
  errcode_t         rc;
  
  rc = ext2fs_read_inode(fs, ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("readlinke: read inode", rc);

  if (!LINUX_S_ISLNK(inode.i_mode))
    return rtems_ext2fs_remap_error("readlinke: not a link", EINVAL);

  if (ext2fs_inode_data_blocks(fs, &inode))
  {
    block = malloc(EXT2_BLOCK_SIZE(fs->super));
    if (!block)
      return rtems_ext2fs_remap_error("readlinke: no memory", ENOMEM);

    rc = io_channel_read_blk(fs->io, inode.i_block[0], 1, block);
    if (rc)
    {
      free(block);
      return rtems_ext2fs_remap_error("readlinke: read block", rc);
    }
    name = (char*) block;
  }
  else
    name = (char *) &(inode.i_block[0]);

  length = strlen(name);
  length = length < (bufsize - 1) ? length : bufsize - 1;
  memcpy(buf, name, length);
  buf[length] = '\0';

  if (block)
    free(block);

  return length;
}

/*
 * rtems_ext2fs_fchmod
 *
 * File change mode routine.
 */
int rtems_ext2fs_fchmod(
  rtems_filesystem_location_info_t *loc,
  mode_t                            mode
)
{
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(loc);
  struct ext2_inode inode;
#if defined(RTEMS_POSIX_API)
  uid_t             uid;
#endif
  errcode_t         rc;

  rc = ext2fs_read_inode(rtems_ext2fs_dev(loc), ino, &inode);
  if (rc)
    return rtems_ext2fs_remap_error("fchmod: read inode", rc);

  /*
   *  Verify I am the owner of the node or the super user.
   */
#if defined(RTEMS_POSIX_API)
  uid = geteuid();

  if ((uid != inode.i_uid) && (uid != 0))
    return rtems_ext2fs_remap_error("fchmod: not owner", EPERM);
#endif

  inode.i_mode &= ~(S_IRWXU | S_IRWXG | S_IRWXO | S_ISUID | S_ISGID | S_ISVTX);
  inode.i_mode |= mode & (S_IRWXU | S_IRWXG | S_IRWXO | S_ISUID | S_ISGID | S_ISVTX);
	inode.i_ctime = time(NULL);

	rc = ext2fs_write_inode(rtems_ext2fs_dev(loc), ino, &inode);
	if (rc)
    return rtems_ext2fs_remap_error("fchmod: write inode", rc);

  rc = ext2fs_write_bitmaps(rtems_ext2fs_dev(loc));
	if (rc)
    return rtems_ext2fs_remap_error("fchmod: write bitmaps", rc);
  
  return 0;
}

/*
 * rtems_ext2fs_fcntl
 *
 * The following routine does a fcntl on an node.
 */

int rtems_ext2fs_fcntl(
  int            cmd,
  rtems_libio_t *iop
)
{
  return 0;
}

/*
 * rtems_ext2fs_stat
 *
 * The following routine does a stat on a node.
 */

int rtems_ext2fs_stat(
  rtems_filesystem_location_info_t *loc,
  struct stat                      *buf
)
{
  ext2_filsys       fs = rtems_ext2fs_dev(loc);
  ext2_ino_t        ino = rtems_ext2fs_get_pathloc_ino(loc);
  struct ext2_inode inode;
  errcode_t         rc;

  memset(buf, 0, sizeof(*buf));

  rc = ext2fs_read_inode(fs, ino, &inode);
  if (rc) 
    return rtems_ext2fs_remap_error("stat: read inode", rc);

   if (LINUX_S_ISCHR(inode.i_mode) || LINUX_S_ISBLK(inode.i_mode))
   {
    int major;
    int minor;
    if (inode.i_block[0])
    {
      major = (inode.i_block[0] >> 8) & 255;
			minor = inode.i_block[0] & 255;
    }
    else
    {
      major = (inode.i_block[1] & 0xfff00) >> 8;
      minor = ((inode.i_block[1] & 0xff) |
               ((inode.i_block[1] >> 12) & 0xfff00));
    }
    buf->st_rdev = rtems_filesystem_make_dev_t(major, minor);
  }

  buf->st_dev     = rtems_ext2fs_blockio_dev(fs);
  buf->st_ino     = ino;
  buf->st_mode    = rtems_ext2fs_modeflags(inode.i_mode);
  buf->st_nlink   = inode.i_links_count;
  buf->st_uid     = inode_uid(inode);
  buf->st_gid     = inode_gid(inode);
  buf->st_atime   = inode.i_atime;
  buf->st_mtime   = inode.i_mtime;
  buf->st_ctime   = inode.i_ctime;
  buf->st_size    = EXT2_I_SIZE(&inode);
  buf->st_blksize = EXT2_BLOCK_SIZE(fs->super);
  buf->st_blocks  = inode.i_blocks;

  return 0;
}

/*
 *  rtems_ext2fs_mknod
 *
 *  Routine to create a node in the EXT2 file system.
 */

int rtems_ext2fs_mknod(
  const char                       *token,      /* IN */
  mode_t                            mode,       /* IN */
  dev_t                             dev,        /* IN */
  rtems_filesystem_location_info_t *pathloc     /* IN/OUT */
)
{
  ext2_filsys       fs = rtems_ext2fs_dev(pathloc);
  ext2_ino_t        parent_ino = rtems_ext2fs_get_pathloc_ino(pathloc);
  struct ext2_inode parent_inode;
  ext2_ino_t        ino;
  struct ext2_inode inode;
  ext2_ino_t        scratch_ino;
  blk_t             blk;
  char*             block = NULL;
  int               flags;
  errcode_t         rc;
  int               imode = rtems_ext2fs_imodeflags(mode);

#if RTEMS_EXT2FS_DEBUG_MKNOD
  {
    const char* type = "unknown";
    if (S_ISDIR(mode))
      type = "dir";
    else if (S_ISCHR(mode))
      type = "char";
    else if (S_ISBLK(mode))
      type = "block";
    else if (S_ISREG(mode))
      type = "file";
    printf("rtems-ext2fs: mknod: parent:%d name:%s type:%s\n",
           parent_ino, token, type);
  }
#endif
  
  rc = ext2fs_new_inode(fs, parent_ino, imode, 0, &ino);
  if (rc)
    return rtems_ext2fs_remap_error("mknod: new inode", ENOMEM);

#if RTEMS_EXT2FS_DEBUG_MKNOD
  printf("rtems-ext2fs: mknod: new ino: %i\n", ino);
#endif
  
  if (parent_ino != ino)
  {
    rc = ext2fs_read_inode(fs, parent_ino, &parent_inode);
    if (rc)
      goto cleanup;
  }
  else
    memset(&parent_inode, 0, sizeof(parent_inode));

  memset(&inode, 0, sizeof(inode));
  inode.i_mode = imode;
#if defined(RTEMS_POSIX_API)
  inode.i_uid = geteuid();
  inode.i_gid = getegid();
#else
  inode.i_uid = 0;
  inode.i_gid = 0;
#endif
  inode.i_atime = inode.i_ctime = inode.i_mtime = time(NULL);;
  inode.i_links_count = 1;
  
  if (S_ISDIR(mode))
  {
    rc = ext2fs_new_block(fs, 0, 0, &blk);
    if (rc)
      goto cleanup;

    rc = ext2fs_new_dir_block(fs, ino, parent_ino, &block);
    if (rc)
		goto cleanup;

    ext2fs_iblk_set(fs, &inode, 1);
    inode.i_block[0] = blk;
    inode.i_links_count = 2;
    inode.i_size = fs->blocksize;

    rc = ext2fs_write_dir_block(fs, blk, block);
    if (rc)
      goto cleanup;

    flags = EXT2_FT_DIR;
  }
  else if (S_ISCHR(mode) || S_ISBLK(mode))
  {
    int major;
    int minor;
    rtems_filesystem_split_dev_t(dev, major, minor);

    if ((major < 256) && (minor < 256))
      inode.i_block[0] = (major * 256) + minor;
    else
      /* The 32bit encoding is [mm mM MM mm]. */
      inode.i_block[1] = (major << 8) | ((minor & ~0xff) << 12) | (minor & 0xff);

    flags = S_ISCHR(mode) ? EXT2_FT_CHRDEV : EXT2_FT_BLKDEV;
  }
  else if (S_ISREG(mode))
  {
    flags = EXT2_FT_REG_FILE;
  }
  else
  {
    rc = EINVAL;
    goto cleanup;
  }

	rc = ext2fs_write_new_inode(fs, ino, &inode);
	if (rc)
		goto cleanup;

  if (ext2fs_test_inode_bitmap(fs->inode_map, ino))
		printf("rtems-ext2fs: mknod: warning: inode already set: %d\n", ino);
  
  rc = ext2fs_lookup(fs, parent_ino, token, strlen(token), 0,
                     &scratch_ino);
  if (!rc)
  {
    rc = EXT2_ET_DIR_EXISTS;
    goto cleanup;
  }

	if (rc != EXT2_ET_FILE_NOT_FOUND)
    goto cleanup;
  
  rc = ext2fs_link(fs, parent_ino, token, ino, flags);
  if (rc == EXT2_ET_DIR_NO_SPACE)
  {
#if RTEMS_EXT2FS_DEBUG_MKNOD
    printf("rtems-ext2fs: mknod: expand dir ino: %i\n", parent_ino);
#endif
    rc = ext2fs_expand_dir(fs, parent_ino);
    if (rc)
      goto cleanup;
    if (parent_ino != ino)
    {
      rc = ext2fs_read_inode(fs, parent_ino, &parent_inode);
      if (rc)
        goto cleanup;
    }
    rc = ext2fs_link(fs, parent_ino, token, ino, flags);
    if (rc)
      goto cleanup;
  }

  if (parent_ino != ino)
  {
    parent_inode.i_links_count++;
    rc = ext2fs_write_inode(fs, parent_ino, &parent_inode);
		if (rc)
			goto cleanup;
  }
    
  if (S_ISDIR(mode))
    ext2fs_block_alloc_stats(fs, blk, +1);
  
	ext2fs_inode_alloc_stats2(fs, ino, +1, 1);

	ext2fs_mark_inode_bitmap(fs->inode_map, ino);
	ext2fs_mark_ib_dirty(fs);
  
  rc = ext2fs_write_bitmaps(fs);
  
cleanup:
  if (block)
    ext2fs_free_mem(&block);
  return rtems_ext2fs_remap_error("mknod:", rc);
}

/*
 *  rtems_ext2fs_rmnod
 *
 *  Routine to remove a node from the EXT2 file system.
 */

int rtems_ext2fs_rmnod(
  rtems_filesystem_location_info_t *parent_pathloc, /* IN */
  rtems_filesystem_location_info_t *pathloc         /* IN */
)
{
  return 0;
}

/*
 * rtems_ext2fs_fdatasync
 *
 * The following routine does a sync on an inode node. Currently it flushes
 * everything.
 */
int rtems_ext2fs_fdatasync(
  rtems_libio_t *iop
)
{
  errcode_t rc;
  
  rc = ext2fs_flush(rtems_ext2fs_dev(&iop->pathinfo));
  if (rc)
    return rtems_ext2fs_remap_error("fdatasync: flush", rc);

  return 0;
}

/*
 *  Handler table for EXT2 link nodes
 */

const rtems_filesystem_file_handlers_r rtems_ext2fs_link_handlers =
{
  .open_h      = NULL,
  .close_h     = NULL,
  .read_h      = NULL,
  .write_h     = NULL,
  .ioctl_h     = NULL,
  .lseek_h     = NULL,
  .fstat_h     = rtems_ext2fs_stat,
  .fchmod_h    = NULL,
  .ftruncate_h = NULL,
  .fpathconf_h = NULL,
  .fsync_h     = NULL,
  .fdatasync_h = NULL,
  .fcntl_h     = NULL,
  .rmnod_h     = rtems_ext2fs_rmnod
};

/*
 * rtems_ext2fs_initialize
 *
 * Forward decl for the ops table.
 */

int rtems_ext2fs_initialize(rtems_filesystem_mount_table_entry_t *mt_entry);

/*
 *  EXT2FS file system operations table
 */

const rtems_filesystem_operations_table rtems_ext2fs_ops =
{
  .evalpath_h     = rtems_ext2fs_eval_path,
  .evalformake_h  = rtems_ext2fs_eval_for_make,
  .link_h         = rtems_ext2fs_link,
  .unlink_h       = rtems_ext2fs_unlink,
  .node_type_h    = rtems_ext2fs_node_type,
  .mknod_h        = rtems_ext2fs_mknod,
  .chown_h        = rtems_ext2fs_chown,
  .freenod_h      = rtems_ext2fs_freenodinfo,
  .mount_h        = NULL,
  .fsmount_me_h   = rtems_ext2fs_initialize,
  .unmount_h      = NULL,
  .fsunmount_me_h = rtems_ext2fs_shutdown,
  .utime_h        = rtems_ext2fs_utime,
  .eval_link_h    = NULL, /* never called cause we lie in the node type */
  .symlink_h      = rtems_ext2fs_symlink,
  .readlink_h     = rtems_ext2fs_readlink
};

/*
 *  rtems_ext2fs_initialize
 */

int rtems_ext2fs_initialize(
  rtems_filesystem_mount_table_entry_t *mt_entry
)
{
  ext2_filsys fs;
  errcode_t   rc;

  initialize_ext2_error_table();
  
  rc = ext2fs_open(mt_entry->dev, EXT2_FLAG_RW, 0, 0,
                   rtems_ext2fs_blockio_manager, &fs);
  if (rc)
    return rtems_ext2fs_remap_error("initialise: open", rc);

  rc = ext2fs_read_bitmaps(fs);
  if (rc)
  {
    ext2fs_close(fs);
    return rtems_ext2fs_remap_error("initialise: read bitmaps", rc);
  }

  mt_entry->fs_info = fs;

  mt_entry->mt_fs_root.node_access = (void*) EXT2_ROOT_INO;
  mt_entry->mt_fs_root.handlers    = &rtems_ext2fs_dir_handlers;
  mt_entry->mt_fs_root.ops         = &rtems_ext2fs_ops;

  return 0;
}
